#ifndef optrar_argh_hh
#define optrar_argh_hh

/* Copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/) */

#include <string>
using std::string;
class ParamHandler
{
private:
    /* No copying */
    void operator=(const ParamHandler &);
    ParamHandler (const ParamHandler &);
    class Reference;
public:
    ParamHandler();
    virtual ~ParamHandler();
    virtual void PrintOpt(unsigned space, const string &opts, const string &desc);
    
    Reference AddLong(const string &longname, int alias);
    Reference AddBool(int c);
    Reference AddInt(int c, int min, int max);
    Reference AddFloat(int c, double min, double max);
    Reference AddString(int c, unsigned min=1, unsigned max=string::npos);
    Reference AddDesc(int c, const string &s, const string &param="");

    void StartParse(int ac, const char *const *av, int firstarg=1);
    const string ProgName() const { return A0; }

    /* --not-x, --no-x, --without-x and --with-x are recognized.
     * Therefore you should use GetBool().
     */
    inline const bool GetBool() const { return polarity; }
    inline const int GetInt() const { return intparm; }
    inline const double GetFloat() const { return doubleparm; }
    inline const string &GetString() const { return param; }
    inline const bool ok() const { return !error; }
    
    int GetParam();
    void ListOptions();
    
private:
    class ArgInfo { public: char type; int min,max;unsigned maxl;double fmin,fmax; };
    class argh_aliasmap *aliases;
    class argh_descsmap *descs;
    class argh_atypemap *argtypes;
    class Reference
    {
        ParamHandler *par;
        int key;
    public:
        
        Reference(ParamHandler *p, int k) : par(p), key(k) { }
        inline Reference &SetBool() { par->AddBool(key); return *this; }
        inline Reference &SetInt(int min, int max) { par->AddInt(key, min, max); return *this; }
        inline Reference &SetFloat(double min, double max) { par->AddFloat(key, min, max); return *this; }
        inline Reference &SetString(unsigned min=1, unsigned max=string::npos) { par->AddString(key, min, max); return *this; }
        inline Reference &SetDesc(const string &s, const string &param="") { par->AddDesc(key, s, param); return *this; }
    };
    Reference MakeRef(int key) { return Reference(this, key); }
    int ParseError() { error=true; return -1; }
    int ShortOpt(int key, const char *s);
private:
    int argc; const char *const *argv;
    const char *A0;
    /* parser volatile */
    int argpos; bool opts; string longo, longobuf; string param;
    const char *shortpointer; bool error, polarity; int intparm; double doubleparm;
};

/*
 * C++ USAGE EXAMPLE
 * 

#include "argh.hh"

int main(int argc, const char *const *argv)
{
    vector<string> files;
    
    string buildfn;
    bool worst = false;
    
    ParamHandler Argh;
    
    Argh.AddLong("worst",   'w').SetBool().SetDesc("Least optimal.");
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("build",   'b').SetString().SetDesc("Builds the archive.", "<file>");
    Argh.AddLong("make",    'b');
    
    Argh.AddBool('k');
    Argh.AddDesc('k', "This option does nothing.");
    Argh.AddLong("idle", 500).SetDesc("This option does nothing.");
    
    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'w': worst = Argh.GetBool(); break;
            case 'V': printf("%s\n", VERSION); return 0;
            case 'b': buildfn = Argh.GetString(); break;
            case 'k': break; // -k
            case 500: break; // --idle
            case 'h':
                printf(
                    "This is software\n"
                    "\nUsage: software [<option> [<...>]] <file> [<...>]\n"
                    "\nThis software does something for the files.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf("\nNo warranty whatsoever.\n");
                return 0;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    if(!files.size())
    {
        fprintf(stderr, "Error: At least one file must be specified.\n");
        return -1;
    }
    
    ...
}

*/

#endif
