#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <linux/kd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <signal.h>
#include <netdb.h>

#include "v5.h"
#include "chrtab.h"

static int Scales[V5_COLS];
static double Res[V5_COLS];
static int Herz[V5_COLS];
static double Age[V5_COLS];

static int fd;
static int Sock;

#define beep 0

/* hsv_to_rgb:
 *  Converts from HSV colorspace to RGB values.
 */
static void hsv_to_rgb(double h, double s, double v, int *r, int *g, int *b)
{
	double f, x, y, z;
	int i;

	v *= 255.0;

	if (s == 0.0)
		*r = *g = *b = (int)v;
	else
	{
		while (h < 0)h += 360;
		h = fmod(h, 360.0) / 60.0;
		i = (int)h;
		f = h - i;
		x = v * (1.0 - s);
		y = v * (1.0 - (s * f));
		z = v * (1.0 - (s * (1.0 - f)));

		switch (i) 
		{
			case 0: *r = v; *g = z; *b = x; break;
			case 1: *r = y; *g = v; *b = x; break;
			case 2: *r = x; *g = v; *b = z; break;
			case 3: *r = x; *g = y; *b = v; break;
			case 4: *r = z; *g = x; *b = v; break;
			case 5: *r = v; *g = x; *b = y; break;
		}
	}
}

#if 0
/* Returns the socket if success.
 * If the connection fails because of connection problems,
 * returns -1. Otherwise (system errors) it fails and terminates.
 */
static int ConnectOutTimed;
static void SubConnectOut(int signum)
{
	signum = ConnectOutTimed++;
}
static int ConnectOut(const char *Server, int Port)
{
    struct hostent *HostName;
    struct sockaddr_in SocketName;
    int plug, p;
    
    if((plug = socket(AF_INET,SOCK_STREAM,0)) < 0)
    	return -1;
    
    HostName = gethostbyname(Server);
    if(!HostName)
    	return -1;
    
    SocketName.sin_family = AF_INET;
    SocketName.sin_port = htons(Port);
    memcpy(&SocketName.sin_addr, HostName->h_addr, HostName->h_length);
	
    signal(SIGABRT, SubConnectOut);
	
    p = fork();
    if(p < 0)
    	return -1;
	
    ConnectOutTimed = 0;
    /* This timeout implementation doesn't eat tons of
     * CPU time looping and doesn't clobber queued alarms.
     * It, however, resets the SIGABRT signal handler. */
    if(!p)
    {
        sleep(5);
        kill(getppid(),SIGABRT);
        _exit(EXIT_SUCCESS);
    } 
    for(;;)
    {
        if((connect(plug,(struct sockaddr *)&SocketName,sizeof(SocketName))) < 0)
        {
        	/* If ConnectOutTimed>0, then errno can contain whatever. */
        	if(errno == EINTR && !ConnectOutTimed)continue;
            if(ConnectOutTimed)errno = ETIMEDOUT;
            close(plug);
            plug = -1;
        }        
        else
        {
        	signal(SIGPIPE, SIG_IGN);
        }
        kill(p, SIGKILL);
        signal(SIGABRT, SIG_DFL);
        waitpid(p, NULL, 0);
        return plug;
    }
}
#endif

static void Init(void)
{
	int a;
	double p = V5_MINHERZ;
	
	MkTab();
	
	fd = open("/dev/console", O_RDWR);
	
	for(a=0; a < V5_COLS; a++)
	{
		Scales[a] = (int)V5_HZPOS(p);
		Herz[a] = (int)p;
		p *= V5_MULTIPLIER;
		
		Age[a] = 100;
	}
	
	Sock = -1;
	
	printf("\33(U");
}

static void Done(void)
{
	printf("\33(B");
	
	if(Sock >= 0)
	{
		close(Sock);
		Sock = -1;
	}
	
	if(fd > 0)
		close(fd);
}

static void Beeppah(int Hz)
{
	if(fd > 0)
		ioctl(fd, KIOCSOUND, Hz ? 1193180/Hz : 0);
}

static void CheckAge(void)
{
	static const char *EfPos = NULL;
	static const char * const Mask[] =
	{
		"........x8x..x7x......\0" ",,,!!74-,",
		".....x88777666xx......\0" "01245689ABGhjzjhGA51-!\15!-",
		".......x7x...x8x......\0" "//,((",
		NULL
	};
	int a;
	
	if(EfPos)
	{
		int c = *EfPos++;
		Beeppah(c == '' ? 0 : c*80);
		if(!*EfPos)EfPos = NULL;
		return;
	}
	
	Beeppah(0);
	
	for(a=30; a<V5_COLS; a++)
	{
		int b;
		const char * const *t;
		for(t=Mask; *t; t++)
		{
			const char *s = *t;
			for(b=0; *s; b++, s++)
			{
				int c1=40,c2=200;
				switch(*s)
				{
					case '5': c1=11; c2=23; break;
					case '6': c1=9; c2=17; break;
					case '7': c1=6; c2=13; break;
					case '8': c1=2; c2=7; break;
					
					case 'x': c1=0; break;	/* Don't care */
				}
				if(a+b >= V5_COLS || Age[a+b]<c1 || Age[a+b]>c2)break;
			}
			if(!*s)
			{
				printf("Effect");
				EfPos = s+1;
				return;
			}
		}
	}
}
static void View(void)
{
	int a, ch, ednyt=Scales[0];
	double Tot, max, toti;
#if beep
	int b=0;
	double bst=0;
#endif

	static int steppi=0;
//	if(!steppi)memset(Res, 0, sizeof Res);
	
	for(max=0, a=0; a<V5_COLS; a++)
	{
		register int dy=0, nyt = Scales[a];
		register double temp = 0;
		
		if(nyt>ednyt)ednyt++;
		for(ch=ednyt; ch<=nyt; ch++)
		{
			dy++;
			temp += Power[ch]-Avg[ch]/AvgCount;
		}
		temp /= dy;
		temp = temp<0 ? 0 : sqrt(temp);
		Res[a] = temp;
		ednyt = nyt;
		
		if(temp>max)max=temp;
	}
	
	max = 3.4; /* /= 1.7;
	
	if(max < 0.8)max = 0.8; */
	
	for(Tot=0, a=0; a<V5_COLS; a++)
		Tot += (Res[a] /= max);
	
	steppi = !steppi;
//	if(steppi)return;

//	Tot /= V5_COLS;

	/* Give quality. */
	toti = Tot;
	if(Tot < 15)Tot = 1E6; 
	
	for(a=0; a<V5_COLS; a++)
	{
		int R,G,B;
		register double S, temp=Res[a], Temp;
		register double lt=a>0         ?Res[a-1]:0;
		register double rt=a<V5_COLS-1 ?Res[a+1]:0;
		
#if beep
		if(temp>0.6 && a>b && a>5)
			b=a, bst=temp;
#endif
		
		Temp = (temp<lt || temp<rt) ? temp/2 : temp;
			
#if 0
		if(temp > Tot/6)
		{
			temp = 10;
			R = 255;
			G = 220;
			B = 150;
		}
		else
#endif
		{
			temp -= 0.2;
		
			S = 1.3-temp;
			if(S > 1)S = 1;
			if(S < 0)S = 0;
				
			if(temp > 1)temp = 1;
			if(temp < 0)temp = 0;
			
			hsv_to_rgb(/*(V5_COLS-a)*360/V5_COLS+*/temp*80, S,temp,  &R,&G,&B);
		}
		
		R = R * (rmax-1) / 255;
		G = G * (gmax-1) / 255;
		B = B * (bmax-1) / 255;
		
     	SetAttr(Lookup[R][G][B] >> 8);
		FlushSetAttr();
		
		putchar(Lookup[R][G][B] & 255);
		
		if(Temp > 0.7)Age[a] = 0;
		else if(Age[a] < 100)Age[a]++;
	}
		
	SetAttr(7);
	FlushSetAttr();	
	
	putchar('|');
	
	CheckAge();
	
	SetAttr(7);
	FlushSetAttr();
	
	#if 0
	if(Sock < 0 && steppi)
	{
		puts("Trying");
		Sock = ConnectOut("10.0.0.4", 9875);
	}
	else if(steppi)
	{
		static int v2=0,v = 0;
		int vol, Vol;
		vol = (toti-12) * 100 / 20;
		Vol = (vol+v) / 2;
		v = vol;
		 
		if(Vol >100)Vol =100;
		if(Vol <  2)Vol = 2;
		
		printf(" %d ", Vol);
		
		if(abs(Vol-v2) > 3)
		{
			if(write(Sock, &Vol, sizeof(Vol)) != 4)
			{
				puts("Write failed.");
				close(Sock);
				Sock = -1;
			}
			v2 = Vol;
		}
	}
	#endif
	puts("");
	
#if beep
	Beeppah(b ? Herz[b] : 0);
#endif
}

Viewer V5 = {fb_not_required, 0, Init, Done, View};
